/***************************************************
** Sagem 9xx Bootstrap Loader v0.9.6-pre2         **
** Copyright (C) 2001 by SoWWWa <sowwwa@atlas.cz> **
**                                                **
** Many thanks to ThS9 for info about             **
** the bootstrap mode                             **
**                                                **
** prerelease version - ONLY FOR TESTING !        **
***************************************************/

#include <stdio.h>
#include <linux/termios.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "sagem.h"
#include "comm.h"
#include "crypt.h"

#define DEFAULT_DEVICE "/dev/ttyS1"
#define DEFAULT_RATE B19200
#define DEFAULT_RATE_LISTEN B57600
#define DEFAULT_BOARD 41
#define DEFAULT_BASE_X41 0x00004000
#define DEFAULT_BASE_X42 0x03001000

int board, outfile, poweron;
dword base, listen_rate, rate;
char *fname, *device, *outfname;

static struct option opts[] = {
	{ "address", 1, NULL, 'a' },
	{ "device", 1, NULL, 'd' },
	{ "listen", 1, NULL, 'l' },
	{ "rate", 1, NULL, 'r' },
	{ "listen-rate", 1, NULL, 'R' },
	{ "power-on", 0, NULL, 'p' },
	{ "board", 1, NULL, 'x' },
	{ "help", 0, NULL, '?' },
	{ "version", 0, NULL, 'V' }
};

void parse_options(int argc, char **argv)
{
	int c, opt_base;
	
	fname = NULL;
	board = DEFAULT_BOARD;
	opt_base = 0;
	device = DEFAULT_DEVICE;
	rate = DEFAULT_RATE;
	listen_rate = DEFAULT_RATE_LISTEN;
	outfname = NULL;
	poweron = 0;
	while ((c = getopt_long(argc, argv, "a:d:l:pr:R:x:?V", opts, NULL)) != -1) {
		switch (c) {
		case 'a':
			base = strtol(optarg, NULL, 0);
			opt_base = 1;
			break;
		case 'd': device = optarg; break;
		case 'l': outfname = optarg; break;
		case 'p': poweron++; break;
		case 'r':
			switch (strtol(optarg, NULL, 0)) {
			case 2400: rate = B2400; break;
			case 4800: rate = B4800; break;
			case 9600: rate = B9600; break;
			case 19200: rate = B19200; break;
			}
			break;
		case 'R':
			switch (strtol(optarg, NULL, 0)) {
			case 2400: listen_rate = B2400; break;
			case 4800: listen_rate = B4800; break;
			case 9600: listen_rate = B9600; break;
			case 19200: listen_rate = B19200; break;
			case 38400: listen_rate = B38400; break;
			case 57600: listen_rate = B57600; break;
			case 115200: listen_rate = B115200; break;
			}
			break;
		case 'x': board = strtol(optarg, NULL, 0); break;
		case '?':
			printf("Sagem 9xx Bootstrap Loader v0.9.6-pre2\n");
			printf("Copyright (C) 2001 by SoWWWa <sowwwa@atlas.cz>\n\n");
			printf("Usage: %s [options] file\n", argv[0]);
			printf("Options:\n");
			printf("  -a BASE       load the code at address BASE\n");
			printf("  -d DEVICE     use serial device DEVICE\n");
			printf("  -l FILE       listen for data after bootstrap and write them into FILE\n");
			printf("  -p            enter the bootstrap mode during power-on\n");
			printf("  -r BAUD       set baudrate to BAUD\n");
			printf("  -R BAUD       set baudrate to BAUD before listening for data\n");
			printf("  -x BOARD      choose board type (BOARD is 41 or 42)\n");
			printf("  -V            print version information\n");
			printf("  -?            print this help\n\n");
			break;
		case 'V':
			printf("Sagem 9xx Bootstrap Loader v0.9.6-pre2\n");
			break;
		}
	}
	if (!opt_base)
		switch (board) {
		case 41: base = DEFAULT_BASE_X41; break;
		case 42: base = DEFAULT_BASE_X42; break;
		}
	if (optind < argc) fname = argv[optind];
}

int kbhit()
{
	fd_set rfds;
	struct timeval tv;
	
	FD_ZERO(&rfds);
	FD_SET(0, &rfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	return select(1, &rfds, NULL, NULL, &tv);
}

void data_listen()
{
	word tmp;
	
	config_serial(listen_rate);
	printf("Waiting for data, writing them into \"%s\"\n", outfname);
	printf("Quit by pressing the enter key...\n");
	while (!kbhit()) {
		recv_byte_wait(&tmp);
		if (tmp != 0x100) write(outfile, &tmp, 1);
		fflush(NULL);
	}
}

void boot_x41(byte *code, dword code_size, dword base)
{
	dword chksum, chk_all, i, j, block_size, seg_size;
	byte cmd[256], resp[256], size, type, shift;
	byte cmd_boot_enable[] = { 0x00, 0x16, 0x00, 0x00, 0x02, 0x00, 0xa5, 0x5a };
	byte cmd_restart[] = { 0x00, 0x16, 0x00, 0x00, 0x00, 0x82, 0x11 };

	printf("Base address: 0x%04lX\n", base);

	config_serial(rate);
	cmd[0] = 0x00; cmd[1] = 0x79;
	send_packet(2, 0x05, cmd);
	recv_packet(&size, &type, resp);
	recv_packet(&size, &type, resp);
	cmd[0] = 0x00; cmd[1] = 0x7a;
	memcpy(cmd + 2, resp + 2, 0x80);
	boot_crypt(0, cmd + 2);
	send_packet(0x82, 0x05, cmd);
	send_packet(sizeof(cmd_boot_enable), 0x05, cmd_boot_enable);
	recv_packet(&size, &type, resp);
	send_packet(sizeof(cmd_restart), 0x05, cmd_restart);
	usleep(100000);

	config_serial(B57600);
	do
		recv_boot_packet(&size, &type, resp);
	while (type != 0x54 || resp[0] != 0x10);

	chk_all = 0;
	for (i = 0; i <= (code_size - 1) / 0x1000; i++, base += 0x1000) {
		chksum = 0;
		shift = 0;

		seg_size = code_size - (i << 12);
		if (seg_size > 0x1000) seg_size = 0x1000;
		printf("Sending 0x20 command (segment address = 0x%04lX, size = 0x%04lX)\n",
			base, seg_size);
		cmd[0] = 0x20;
		memcpy(cmd + 1, &base, 4);
		memcpy(cmd + 5, &seg_size, 2);
		boot_chksum(cmd + 1, 6, &chksum, &shift);
		send_packet(7, 0x54, cmd);
		recv_boot_packet(&size, &type, resp);

		boot_chksum(code + (i << 12), seg_size, &chksum, &shift);
		cmd[0] = 0x21;
		for (j = 0; j <= (seg_size - 1) / 0xf0; j++) {
			block_size = seg_size - j * 0xf0;
			if (block_size > 0xf0) block_size = 0xf0;
			printf("Sending 0x21 command (0x%lX bytes)\n", block_size);
			memcpy(cmd + 1, code + (i << 12) + j * 0xf0, block_size);
			send_packet(block_size + 1, 0x54, cmd);
			recv_boot_packet(&size, &type, resp);
		}

		chk_all ^= chksum;

		printf("Sending 0x22 command\n");
		cmd[0] = 0x22;
		memcpy(cmd + 1, &chksum, 4);
		send_packet(5, 0x54, cmd);
		recv_boot_packet(&size, &type, resp);
	}

	printf("Sending 0x23 command\n");
	cmd[0] = 0x23;
	memcpy(cmd + 1, &chk_all, 4);
	boot_chksum_all(code, code_size, (word *) (cmd + 5));
	send_packet(7, 0x54, cmd);
	recv_boot_packet(&size, &type, resp);
	
	printf("Bootstrap code started\n");
	if (outfile != -1) data_listen();
}

void boot_x42(byte *code, dword code_size, dword base)
{
	byte cmd[256], resp[256], size, type;
	byte cmd_boot_enable[] = { 0x00, 0x16, 0x03, 0x00, 0x00, 0x00, 0x5a, 0xa5 };
	byte cmd_restart[] = { 0x00, 0x89 };

	printf("Base address: 0x%08lX\n\n", base);

	config_serial(rate);
	cmd[0] = 0x00; cmd[1] = 0x79;
	send_packet(2, 0x05, cmd);
	recv_packet(&size, &type, resp);
	recv_packet(&size, &type, resp);
	cmd[0] = 0x00; cmd[1] = 0x7a;
	memcpy(cmd + 2, resp + 2, 0x80);
	boot_crypt(4, cmd + 2);
	send_packet(0x82, 0x05, cmd);
	send_packet(sizeof(cmd_boot_enable), 0x05, cmd_boot_enable);
	recv_packet(&size, &type, resp);
	send_packet(sizeof(cmd_restart), 0x05, cmd_restart);
	usleep(100000);

	config_serial(B57600);
	do
		recv_boot_packet(&size, &type, resp);
	while (type != 0x54 || resp[0] != 0x10);

	send_x42_boot_packet(base + 1, 1, NULL);
	send_x42_boot_packet(base, code_size, code);
	send_x42_boot_packet(0x03000000, 0x00000002, NULL);
	send_x42_boot_packet(0xFFFFFFFF, 0x00000000, NULL);

	if (outfile != -1) data_listen();
}

void boot_x42_poweron(byte *code, dword code_size, dword base)
{
	byte b;
	word w;

	printf("Base address: 0x%08lX\n\n", base);

	config_serial(B57600);
	printf("Sending 0x55 periodically (1ms), waiting for 0xB0\n");
	printf("Now you can turn the phone on...\n<");
	do {
		send_byte_noecho(0x55);
		recv_byte_wait(&w);
	} while (w != 0xB0);
	printf("\n");

	send_x42_boot_packet(base + 1, 1, NULL);
	send_x42_boot_packet(base, code_size, code);
	send_x42_boot_packet(0x03000000, 0x00000002, NULL);
	send_x42_boot_packet(0xFFFFFFFF, 0x00000000, NULL);
	
	printf("Waiting for the acknowledge byte (0xEF)\n<");
	recv_byte(&b);
	if (b == 0xEF)
		printf("\nCode loaded successfully\n");
	else
		printf("\nError while loading code\n");
	
	if (outfile != -1) data_listen();
}

void boot_x51(byte *code, dword code_size, dword base)
{
	dword chksum, chk_all, i, j, block_size, seg_size;
	byte cmd[256], resp[256], size, type, shift;
	byte cmd_boot_enable[] = { 0x00, 0x16, 0x00, 0x00, 0xfd, 0x0e, 0xa5, 0x5a };
	byte cmd_restart[] = { 0x00, 0x89 };

	printf("Base address: 0x%04lX\n", base);

	config_serial(rate);
	cmd[0] = 0x00; cmd[1] = 0x79;
	send_packet(2, 0x05, cmd);
	recv_packet(&size, &type, resp);
	recv_packet(&size, &type, resp);
	cmd[0] = 0x00; cmd[1] = 0x7a;
	memcpy(cmd + 2, resp + 2, 0x80);
	boot_crypt(4, cmd + 2);
	send_packet(0x82, 0x05, cmd);
	send_packet(sizeof(cmd_boot_enable), 0x05, cmd_boot_enable);
	recv_packet(&size, &type, resp);
	send_packet(sizeof(cmd_restart), 0x05, cmd_restart);
	usleep(100000);

	config_serial(B57600);
	do
		recv_boot_packet(&size, &type, resp);
	while (type != 0x54 || resp[0] != 0x10);

	usleep(1000);
	config_serial(B115200);
	printf(">");
	send_block(code, code_size);
	printf("\n");

	if (outfile != -1) data_listen();
}

int main(int argc, char **argv)
{
	int file;
	byte *code;
	dword code_size;

	parse_options(argc, argv);
	if (fname == NULL) return 0;
	if ((file = open(fname, O_RDONLY)) == -1) {
		perror("Bootstrap code could not be read");
		return 1;
	}
	if (outfname != NULL)
		outfile = creat(outfname, 0644);
	else
		outfile = -1;

	code_size = lseek(file, 0, SEEK_END);
	lseek(file, 0, SEEK_SET);
	code = (void *) malloc(code_size);
	read(file, code, code_size);
	printf("Bootstrap code size: 0x%04lX\n", code_size);
	close(file);
		
	if (init_serial(device) == -1) {
		free(code);
		return 1;
	}
	switch (board) {
	case 41: boot_x41(code, code_size, base); break;
	case 42:
		if (poweron)
			boot_x42_poweron(code, code_size, base);
		else
			boot_x42(code, code_size, base);
		break;
	case 51: boot_x51(code, code_size, base); break;
	}
	close_serial();

	free(code);

	return 0;
}
