#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <unistd.h>
#include <linux/serial.h>

#include "sagem.h"

int comm;
byte chksum;

void send_byte(byte x)
{
	chksum ^= x;
	printf(" %02X", x);
	write(comm, &x, 1);
}

void recv_byte(byte * x)
{
	read(comm, x, 1);
	printf(" %02X", *x);
	chksum ^= *x;
}

void send_byte_noecho(byte x)
{
	write(comm, &x, 1);
}

void recv_byte_wait(word * x)
{
	fd_set rfds;
	struct timeval tv;
	
	FD_ZERO(&rfds);
	FD_SET(comm, &rfds);
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	if(select(comm + 1, &rfds, NULL, NULL, &tv)) {
		*x = 0;
		read(comm, x, 1);
		printf(" %02X", *x);
	} else
		*x = 0x100;
}

void send_block(byte * data, dword size)
{
	dword i;

	for (i = 0; i < size; i++)
		send_byte(data == NULL ? 0 : data[i]);
}

void recv_block(byte * data, dword size)
{
	dword i;

	for (i = 0; i < size; i++)
		recv_byte(data + i);
}

void send_packet(byte size, byte type, byte * data)
{
	chksum = 0;

	printf(">");
	send_byte(0x02);
	send_byte(size + 2);
	send_byte(0xD2);
	send_byte(type);
	send_block(data, size);
	send_byte(chksum);
	send_byte(0x03);
	printf("\n");
}

void recv_boot_packet(byte * size, byte * type, byte * data)
{
	byte tmp;

	printf("<");
      retry:
	chksum = 0;
	recv_byte(&tmp);
	if (tmp != 0x02)
		goto retry;
	recv_byte(size);
	*size -= 2;
	recv_byte(&tmp);
	if (tmp != 0xD2)
		goto retry;
	recv_byte(type);
	recv_block(data, *size);
	recv_byte(&tmp);
	if (chksum != 0)
		goto retry;
	recv_byte(&tmp);
	if (tmp != 0x03)
		goto retry;

	send_byte(0x06);
	printf("\n");
}

void recv_packet(byte * size, byte * type, byte * data)
{
	byte tmp;

	printf("<");
      retry:
	chksum = 0;
	recv_byte(&tmp);
	if (tmp != 0x02)
		goto retry;
	recv_byte(size);
	*size -= 2;
	recv_byte(&tmp);
	if (tmp != 0xD2)
		goto retry;
	recv_byte(type);
	recv_block(data, *size);
	recv_byte(&tmp);
	if (chksum != 0)
		goto retry;
	recv_byte(&tmp);
	if (tmp != 0x03)
		goto retry;
	printf("\n");
}

void send_x42_boot_packet(dword addr, dword size, byte * data)
{
	printf(">");
	send_byte(addr >> 24);
	send_byte(addr >> 16);
	send_byte(addr >> 8);
	send_byte(addr);
	send_byte(size >> 24);
	send_byte(size >> 16);
	send_byte(size >> 8);
	send_byte(size);
	send_block(data, size);
	printf("\n");
}

int config_serial(dword speed)
{
	int err;
	struct termios termios;

	if ((err = ioctl(comm, TCGETS, &termios)) == -1) {
		perror("Reading serial port params");
		return -1;
	}
	termios.c_iflag = 0;
	termios.c_oflag = 0;
	termios.c_cflag = CS8 | CREAD | CLOCAL | CSTOPB | (speed & CBAUD);
	termios.c_lflag = FLUSHO;
	if ((err = ioctl(comm, TCSETS, &termios)) == -1) {
		perror("Setting serial port params");
		return -1;
	}
	
	return 0;
}

int init_serial(char * dev)
{
	if ((comm = open(dev, O_RDWR | O_SYNC)) == -1)
		perror("Opening serial port");
	return comm;
}

int close_serial()
{
	int err;
	
	if ((err = close(comm)) == -1)
		perror("Opening serial port");
	return err;
}
